//
//  AppDelegate.h
//  GeckoEmbed
//
//  Created by Ted Mielczarek on 3/2/15.
//  Copyright (c) 2015 Mozilla. All rights reserved.
//

#import <UIKit/UIKit.h>

@interface AppDelegate : UIResponder <UIApplicationDelegate>

@property (strong, nonatomic) UIWindow *window;


@end

