/* -*- Mode: IDL; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#include "nsIDOMElement.idl"

interface nsIRDFCompositeDataSource;
interface nsIRDFResource;
interface nsIControllers;
interface nsIBoxObject;


[uuid(75435ab3-6863-42a1-ade3-025393d9e80e)]
interface nsIDOMXULElement : nsIDOMElement
{
};
